import argparse
import importlib.util
import json

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Driver for running test cases for Lambda functions",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="Enjoy :)",
    )

    parser.add_argument(
        "filename",
        help="Name of the application",
        type=str,
        default="lambda.py",
    )
    parser.add_argument(
        "--handler",
        help="Name of the handler function",
        type=str,
        default="handler",
    )
    parser.add_argument(
        "--test",
        "-t",
        help="Path to the test cases file",
        type=str,
        default="data.json",
    )

    args = parser.parse_args()

    file_path = args.filename
    module_name = file_path.split(".")[0]

    # Find spec from file location
    spec = importlib.util.spec_from_file_location(module_name, file_path)

    # Load the module from the specification
    module = importlib.util.module_from_spec(spec)

    # Execute the module
    spec.loader.exec_module(module)

    handler = getattr(module, args.handler)

    # open data.json file
    with open(args.test) as f:
        data = json.load(f)

        for entry in data["tests"]:
            event = entry["event"]
            context = entry["context"]
            print(handler(event, context))



import argparse

from ltrim.moduify.moduifier import Moduify

__all__ = ["Moduify"]


def main():
    parser = argparse.ArgumentParser(
        description="Modify modules by removing attributes."
    )
    parser.add_argument("module_name", type=str, help="Name of the module to modify.")
    parser.add_argument(
        "-r",
        "--remove",
        action="store_true",
        help="""Remove or keep attributes. If set, attributes
        are removed. Otherwise, they are kept.""",
    )
    parser.add_argument(
        "-a",
        "--attributes",
        type=str,
        nargs="+",
        help="Attributes to remove or keep.",
    )

    args = parser.parse_args()

    moduifier = Moduify(module_name=args.module_name)
    moduifier.modify(args.attributes, remove=args.remove)



# if python versionh is 3.11 or higher, use the sys.stdlib_module_names
import random
import re
import sys
from functools import wraps
from multiprocessing import Pipe, Process

if sys.version_info >= (3, 11):
    blacklist = sys.stdlib_module_names
else:
    blacklist = [
        "ast",
        "argparse",
        "os",
        "sys",
        "shutil",
        "stat",
        "importlib",
        "re",
        "json",
        "time",
    ]


def isolate(func):
    """
    A decorator to run a function in a separate process and send its return value
    through a pipe.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Create a pipe for communication
        parent_conn, child_conn = Pipe()

        def target(pipe_conn, *args, **kwargs):
            # Call the wrapped function and send the result through the pipe
            result = func(*args, **kwargs)
            pipe_conn.send(result)
            pipe_conn.close()  # Close the pipe connection after sending the result

        # Run the function in a separate process, passing the child pipe connection
        process = Process(target=target, args=(child_conn, *args), kwargs=kwargs)
        process.start()
        process.join()  # Wait for the process to complete

        # Return the parent pipe connection for the caller to retrieve the result
        return parent_conn.recv()

    return wrapper


def scoring(method, t, m, T, M):
    """
    The scoring method to calculate the top K ranking of the modules.

    :param method: The scoring method to use
    :param t: The import time of the module
    :param m: The memory usage of the module
    :param T: The total import time of the application
    :param M: The total memory usage of the application
    :return: The score of the module based on the scoring method
    """
    if method == "time":
        return t
    elif method == "memory":
        return m
    elif method == "cost":
        return (T - t) * m + t * M
    elif method == "random":
        return random.random()
    elif method == "custom":
        pass  # Placeholder for custom scoring method
    else:
        raise ValueError(
            "Invalid scoring method. Choose from 'time', 'memory', 'cost' or 'random'."
        )


def sort_report(report, method, T, M):
    """
    Sort the profiling report based on the scoring method.

    :param report: The profiling report to sort
    :param method: The scoring method to use for sorting
    :return: The sorted list of modules based on the scoring method
    """

    # Add scoring to the report
    for entry in report:
        report[entry]["score"] = scoring(
            method,
            report[entry]["time"],
            report[entry]["memory"],
            T,
            M,
        )

    # Step 4: Sort the modules based on the scoring method
    return sorted(report.items(), key=lambda x: x[1]["score"], reverse=True)


def filter_pycg(module, pycg_attributes):
    """
    Filter the PyCG attributes by keeping only the ones that are in the module.

    :param module: The module to filter the attributes
    :param pycg_attributes: The PyCG attributes to filter
    :return: The filtered PyCG attributes
    """

    # using regex to match the module name, i.e. "module.*"
    pattern = re.compile(r"{}.*".format(module))
    return [
        attr.removeprefix(module + ".")
        for attr in pycg_attributes
        if pattern.match(attr)
    ]


def update_alive_modules(alive_modules, report):
    """
    Update the alive modules based on the report. Dictionaries are mutable objects,
    so the alive_modules set will be updated in place.

    :param alive_modules: The set of alive modules
    :param report: The profiling report
    """
    modules_to_remove = [module for module in alive_modules if module not in report]
    for module in modules_to_remove:
        alive_modules.discard(module)



import ast

from ltrim.transformers.utils import retrieve_name


class ImportsFinder(ast.NodeVisitor):
    """
    Find all import statements in a Python module
    """

    def __init__(self):
        ast.NodeVisitor.__init__(self)
        self.imports = []

    def visit_Import(self, node):
        """
        Custom visit_Import
        """
        self.imports.extend(n.name for n in node.names)
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ImportFrom(self, node):
        """
        Custom visit_ImportFrom
        """
        self.imports.extend(node.module for n in node.names)
        ast.NodeVisitor.generic_visit(self, node)


class RemoveAttribute(ast.NodeTransformer):
    """
    Remove attributes from a Python module

    :param attributes: A list of attributes to remove
    """

    def __init__(self, attributes):
        self.attributes = dict(attributes)

    def visit_Assign(self, node):
        """
        Custom visit_Assign
        """

        newnode = node
        if isinstance(node.value, ast.Call):
            s = retrieve_name(node.value.func)
            if s in self.attributes:
                return ast.Pass()

        first_target = node.targets[0]
        if isinstance(first_target, ast.Name) and (first_target.id == "__all__"):
            new_elements = []
            if not hasattr(node.value, "elts"):
                return node
            for element in node.value.elts:
                if retrieve_name(element) in self.attributes:
                    continue
                new_elements.append(element)
            node.value.elts = new_elements

        for target in node.targets:
            if isinstance(target, ast.Name) and target.id in self.attributes:
                if self.attributes[target.id] == "variable":
                    newnode = ast.Pass()

        return newnode

    def visit_FunctionDef(self, node):
        """
        Custom visit_FunctionDef
        """
        if node.name in self.attributes:
            if self.attributes[node.name] == "function":
                return ast.Pass()
        return node

    def visit_ClassDef(self, node):
        """
        Custom visit_ClassDef
        """
        if node.name in self.attributes:
            if self.attributes[node.name] == "class":
                return ast.Pass()
        return node

    def visit_Import(self, node):
        """
        Custom visit_Import
        """
        newnames = []

        for alias in node.names:
            if alias.asname is not None:
                if alias.asname not in self.attributes:
                    newnames.append(alias)
            else:
                if alias.name not in self.attributes:
                    newnames.append(alias)

        if len(newnames) == 0:
            return ast.Pass()

        node.names = newnames

        return node

    def visit_ImportFrom(self, node):
        """
        Custom visit_ImportFrom
        """
        newnames = []

        for alias in node.names:
            if alias.asname is not None:
                if alias.asname not in self.attributes:
                    newnames.append(alias)
            else:
                if alias.name not in self.attributes:
                    newnames.append(alias)

        if len(newnames) == 0:
            return ast.Pass()

        node.names = newnames

        return node


class SetFix(ast.NodeTransformer):
    """
    Fix modules with __all__ = list(set(...)) pattern
    """

    def __init__(self, attributes):
        self.attributes = attributes
        self.fix_lock = False

    def visit_Assign(self, node):
        """
        Custom visit_Assign
        """
        target = node.targets[0]

        if isinstance(target, ast.Name) and target.id == "__all__":
            if isinstance(node.value, ast.Call):
                if isinstance(node.value.func, ast.Name):
                    if node.value.func.id == "list":
                        self.fix_lock = True

        self.generic_visit(node)
        self.fix_lock = False

        return node

    def visit_Call(self, node):
        """
        Custom visit_Call
        """
        if self.fix_lock:
            if isinstance(node.func, ast.Name) and node.func.id == "set":
                new_elements = []
                for element in node.args:
                    c1 = isinstance(element, ast.Attribute)
                    name = retrieve_name(element.value)
                    name = name.split(".")[0] if "." in name else name
                    c3 = name in self.attributes
                    if c1 and c3:
                        continue
                    new_elements.append(element)
                node.args = new_elements
            else:
                self.generic_visit(node)
        else:
            self.generic_visit(node)

        return node



import argparse

from ltrim.delta.delta import DeltaDebugger
from ltrim.delta.utils import PyLambdaRunner

__all__ = ["DeltaDebugger", "PyLambdaRunner"]


def main():
    parser = argparse.ArgumentParser(
        description="""
        Run Delta Debugging on a module with a target program and a
        set of attributes to keep.
        """
    )
    parser.add_argument("target", help="The target program")
    parser.add_argument("module", help="The module to modify")
    parser.add_argument(
        "--test-cases",
        type=str,
        default="data.json",
        help="JSON file containing test cases.",
    )
    parser.add_argument("attributes", type=str, nargs="*", help="Attributes to keep.")

    args = parser.parse_args()

    debugger = DeltaDebugger(
        target=args.target,
        module_name=args.module,
        marked_attributes=args.attributes,
        test_cases=args.test_cases,
    )

    remaining_attributes = debugger.delta_debug()
    print(f"Remaining attributes: {remaining_attributes}")
    debugger.finalize_module(remaining_attributes)



import argparse
import logging

from ltrim.debloat.debloat import Debloater
from ltrim.utils import Config


def main():
    parser = argparse.ArgumentParser(
        prog="debloat",
        description="""Debloat a Python application by removing unused
        attributes of imported modules.""",
        epilog="""Developed and maintained by Spyros Pavlatos. Originally created
        in the Distributed Systems Laboratory at University of Pennsylvania""",
    )

    parser.add_argument("filename", type=str, help="Name of the application")

    parser.add_argument(
        "-k", "--top-K", type=int, default=10, help="Number of modules to debloat."
    )

    parser.add_argument(
        "-t",
        "--testcases",
        type=str,
        default="data.json",
        help="Path to the testcases file.",
    )

    parser.add_argument(
        "-s",
        "--scoring",
        default="cost",
        choices=["cost", "time", "memory", "random", "custom"],
        help="The scoring method to calculate the top K ranking of the modules.",
    )

    parser.add_argument(
        "-H",
        "--handler",
        default="handler",
        help="""The name of the function handler. 
        This should be the entry point of the application""",
    )

    # Disable PyCG flag
    parser.add_argument(
        "--no-pycg",
        action="store_true",
        help="""Do not use PyCG for the debloating
        process. This will result in a slower debloating
        process.""",
    )

    args = parser.parse_args()

    # create a configuration
    config = Config(
        appname=args.filename,
        handler=args.handler,
        test_cases=args.testcases,
    )

    debloater = Debloater(
        config=config,
        top_K=args.top_K,
        scoring=args.scoring,
        disable_pycg=args.no_pycg,
    )

    debloater.run()



import csv
from typing import NewType

DeltaRecord = NewType("DeltaRecord", tuple[float, int, int])


class Stats:
    """
    Class to store statistics about the debloating process

    :param appname: The name of the application
    :param top_K: The number of modules to debloat
    """

    def __init__(self, appname, top_K):
        self.stats = {}
        self.appname = appname.removesuffix(".py")
        self.top_K = top_K

    def add_module(self, module):
        """
        Add a module to the internal dictionary

        :param module: The module to add
        """
        self.stats[module] = ModuleRecord(module)

    def set_profiling_stats(self, module, memory, time, before=True):
        """
        Set stats after profiling

        :param module: The module to set the stats for
        :param memory: Memory (in MB)
        :param time: Time (in ms)
        :param before: Whether the profiling happened before or after DD
        """
        order = "Pre" if before else "Post"
        self.stats[module].set_profiling_stats(memory, time, order)

    def set_debloating_stats(self, module, debloat_record: DeltaRecord):
        """
        Set the debloating stats

        :param module: The module to set the stats for
        :param debloat_record: The debloating stats
        """
        self.stats[module].set_debloating_stats(debloat_record)

    def set_path(self, module, path):
        """
        Set the path to the module file

        :param module: The module to set the stats for
        :param path: The path to the module file
        """
        self.stats[module].set_path(path)

    def convert_to_csv(self):
        """
        Convert the internal dictionary to a CSV
        """

        filename = f"log/{self.appname}_{self.top_K}_stats.csv"
        keys = [
            "Module",
            "Pre Memory",
            "Pre Import Time",
            "Post Memory",
            "Post Import Time",
            "Debloat Time",
            "Pre Attributes",
            "Removed Attributes",
            "Path",
        ]

        with open(filename, mode="w", encoding="utf-8") as file:
            writer = csv.DictWriter(file, fieldnames=keys)
            writer.writeheader()

            for _, record in self.stats.items():
                row = record.convert_to_row()
                writer.writerow(row)


class ModuleRecord:
    """
    Class to store statistics about the debloating process
    of a module

    :param module_name: The module to be debloated
    :param path: The path to the module file
    """

    def __init__(self, module_name):
        self.module_name = module_name
        self.stats = {
            "Pre Memory": 0,
            "Pre Import Time": 0,
            "Post Memory": 0,
            "Post Import Time": 0,
            "Debloat Time": 0,
            "Pre Attributes": 0,
            "Removed Attributes": 0,
            "Path": "",
        }
        self.path = None

    def set_profiling_stats(self, memory, time, order):
        """
        Set stats after profiling

        :param memory: Memory (in MB)
        :param time: Time (in ms)
        :param before: Whether the profiling happened before or after DD
        """
        self.stats[order + " Memory"] = memory
        self.stats[order + " Import Time"] = time

    def set_debloating_stats(self, debloat_record: DeltaRecord):
        """
        Set the debloating stats

        :param debloat_time: Time taken to debloat the module (in ms)
        :param attributes_before: Number of module's attributes before debloating
        :param attributes_after: Number of module's attributes after debloating
        """
        debloat_time, attributes_before, attributes_after = debloat_record

        self.stats["Debloat Time"] = debloat_time
        self.stats["Pre Attributes"] = attributes_before
        self.stats["Removed Attributes"] = attributes_after

    def set_path(self, path):
        """
        Set the path to the module file

        :param path: The path to the module file
        """
        self.path = path

    def convert_to_row(self):
        """
        Convert the internal dictionary to a CSV row
        """

        row = {
            "Module": self.module_name,
            "Pre Memory": self.stats["Pre Memory"],
            "Pre Import Time": self.stats["Pre Import Time"],
            "Post Memory": self.stats["Post Memory"],
            "Post Import Time": self.stats["Post Import Time"],
            "Debloat Time": self.stats["Debloat Time"],
            "Pre Attributes": self.stats["Pre Attributes"],
            "Removed Attributes": self.stats["Removed Attributes"],
            "Path": self.path,
        }

        return row



import ast
import importlib
import logging
import sys
import time

from ltrim.delta.utils import Found, PyLambdaRunner, chunks, flatten
from ltrim.moduify import Moduify
from ltrim.utils import MAGIC_ATTRIBUTES, Config, DeltaRecord, cmd_message, mkdirp


class DeltaDebugger:
    """
    Delta Debugger instance

    :param target: The target program to run
    :param module_name: Name of the module to debloat
    :param marked_attributes: Attributes that must be kept
    """

    def __init__(
        self,
        config: Config,
        module_name,
        marked_attributes,
    ):
        self.config = config
        self.target = config.appname
        # Attributes that must be kept
        self.marked_attrs = marked_attributes

        # Initialize the Moduify instance
        self.module_name = module_name
        self.moduifier = Moduify(
            module_name=self.module_name,
            marked_attributes=self.marked_attrs,
        )

        # Initialize the logger for the module under DD
        self.logger = logging.getLogger(module_name + "_delta")
        self.logger.setLevel(logging.INFO)
        handler = logging.FileHandler(f"log/{module_name}_delta.log")
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.propagate = False

        self.iterations = 0

        # Create a logging directory for intermediate results
        mkdirp("log/" + self.module_name + "/iterations")

        # Instace of driver for running the target program
        self.runner = PyLambdaRunner(config=self.config)

        process = self.runner.run()

        if process.returncode == 0:
            self.original_output = str(process.stdout, "utf-8")
            self.logger.info("Original output: %s", self.original_output)
        else:
            self.logger.error(f"Error running target program: {process.stderr}")
            sys.exit(1)

    def oracle(self, attributes, log=True):
        """
        Run the target program with the modified module and attributes.
        If the program fails to run, the oracle returns False.
        Otherwise, it compares the output of the program with the
        original output.

        :param attributes: The attributes under test
        """

        self.iterations += 1

        if log:
            try:
                modified_ast = self.moduifier.modify(attributes, remove=False)

                iteration_dir = (
                    "log/" + self.module_name + "/iterations/i" + str(self.iterations)
                )
                mkdirp(iteration_dir)

                with open(
                    iteration_dir + "/__init__.py", "w", encoding="utf-8"
                ) as file:
                    new_source = ast.unparse(modified_ast)
                    file.write(new_source)

                with open(iteration_dir + "/attr.txt", "w", encoding="utf-8") as file:
                    file.write("Keeping the following attributes:\n")
                    for item in attributes:
                        file.write(f"{item}\n")

            except Exception as e:
                self.logger.error("Error modifying module: %s", e)
                cmd_message(f"Error modifying module: {e}", "error")

                return False

        process = self.runner.run()

        if process.returncode == 0:
            output = str(process.stdout, "utf-8")

            if output != self.original_output:
                self.logger.info("Output changed: %s", output)
                self.logger.info("Original output: %s", self.original_output)
                return False

            return output == self.original_output

        else:
            self.logger.error("Error running target program: %s", process.stderr)
            return False

    def delta_debug(self, log=False):
        """
        Delta-Debugging algorithm
        """

        start = time.time()

        if self.moduifier.ast is None:
            cmd_message("Module is not a Python file")
            return [], DeltaRecord((0, 0, 0))

        cmd_message("Running Delta Debugging for module " + self.module_name)

        self.logger.info("Running DeltaDebugging for module %s", self.module_name)
        self.logger.info("Necessary attributes: %s", self.marked_attrs)

        module = importlib.import_module(self.module_name)

        members = [x for x in dir(module) if x not in MAGIC_ATTRIBUTES]
        self.logger.info("Module attributes: %s", members)

        remaining_attrs, n = members, 2

        all_attributes = len(dir(module))
        attrs_before = len(remaining_attrs)

        while n <= len(remaining_attrs):
            us = list(chunks(remaining_attrs, n))

            try:
                for i in range(n):
                    attributes = us[i]

                    self.logger.info("Trying partition %s", attributes)

                    if self.oracle(attributes, log):
                        remaining_attrs, n = attributes, 2
                        raise Found

                if n > 2:
                    for i in range(n):
                        coattributes = us.copy()
                        coattributes.pop(i)
                        coattributes = flatten(coattributes)

                        self.logger.info("Trying c-partition %s", coattributes)

                        if self.oracle(coattributes, log):
                            remaining_attrs, n = coattributes, n - 1
                            raise Found

                n *= 2

            except Found:
                self.logger.info("REDUCED to %s", remaining_attrs)
                continue

        end = time.time()
        debloat_time = end - start
        cmd_message(
            f"Total time taken to debloat {self.module_name}: {debloat_time:.2f}s"
        )

        self.logger.info("Remanining attributes: %s", remaining_attrs)

        attrs_after = len(remaining_attrs)
        removed = attrs_before - attrs_after
        cmd_message(
            f"Removed {removed} attributes {(removed / all_attributes * 100):.2f}%."
        )

        delta_record = DeltaRecord((debloat_time, all_attributes, attrs_after))

        return list(self.marked_attrs) + remaining_attrs, delta_record

    def get_attr_stats(self):
        """
        Wrapper around iterations stats
        """
        return self.iterations

    def finalize_module(self, attributes, local=False):
        """
        Finalize the module by removing a set of attributes

        :param attributes: The attributes to remove
        :param local: In local environment, restore the original directory
        """

        m_path = self.moduifier.module_path

        if self.moduifier.ast is None:
            return m_path

        m_ast = self.moduifier.modify(attributes, remove=False)

        basename = self.moduifier.basename
        log_mod_dir = "log/" + self.module_name

        # Log the modified attributes
        attributes_log = log_mod_dir + "/attrs.txt"
        with open(attributes_log, "w", encoding="utf-8") as file:
            for item in attributes:
                file.write(f"{item}\n")
            file.flush()

        # Log the modified and the original __init__.py files
        mod_init_path = log_mod_dir + "/" + basename
        with open(mod_init_path, "w", encoding="utf-8") as file:
            new_source = ast.unparse(m_ast)
            file.write(new_source)
            file.flush()

        with open(log_mod_dir + "/original_" + basename, "w", encoding="utf-8") as file:
            with open(
                self.moduifier.backup_dir + "/" + basename,
                "r",
                encoding="utf-8",
            ) as f:
                file.write(f.read())
                file.flush()

        if local:
            self.moduifier.restore_original_directory()

        # Debloat stub
        with open(m_path, "a") as file:
            file.write("\n# Debloated\n")

        return m_path



import os
import shutil


def cp(source_path, destination_path):
    """
    Copy a file from source to destination.
    Same functionality as: cp src dst

    :param source_path: Source file path
    :param destination_path: Destination file path
    """

    # Create the destination directory if it doesn't exist
    destination_dir = os.path.dirname(destination_path)
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)

    # If the destination file exists, remove it
    if os.path.exists(destination_path):
        os.remove(destination_path)

    # Copy the file to the destination
    shutil.copy2(source_path, destination_path)


def mkdirp(directory_path):
    """
    Same functionality as: mkdir -p dir
    """

    if os.path.exists(directory_path):
        shutil.rmtree(directory_path)
    os.makedirs(directory_path)
